import numpy as np
import torch
from torch import nn
from onnx import numpy_helper

from onnx2pytorch.operations import (
    BatchNormWrapper,
    InstanceNormWrapper,
    LSTMWrapper,
)
from onnx2pytorch.convert.attribute import extract_attributes, extract_attr_values


def extract_params(params):
    """Extract weights and biases."""
    param_length = len(params)
    if param_length == 1:
        weight = params[0]
        bias = None
    elif param_length == 2:
        weight = params[0]
        bias = params[1]
    else:
        raise ValueError("Unexpected number of parameters: {}".format(param_length))
    return weight, bias


def load_params(layer, weight, bias):
    """Load weight and bias to a given layer from onnx format."""
    layer.weight.data = torch.from_numpy(numpy_helper.to_array(weight))
    if bias is not None:
        layer.bias.data = torch.from_numpy(numpy_helper.to_array(bias))


def convert_layer(node, layer_type, params=None):
    """Use to convert Conv, MaxPool, AvgPool layers."""
    assert layer_type in [
        "Conv",
        "ConvTranspose",
        "MaxPool",
        "AvgPool",
    ], "Incorrect layer type: {}".format(layer_type)
    kwargs = extract_attributes(node)
    kernel_size_length = len(kwargs["kernel_size"])
    try:
        layer = getattr(nn, "{}{}d".format(layer_type, kernel_size_length))
    except AttributeError:
        raise ValueError(
            "Unexpected length of kernel_size dimension: {}".format(kernel_size_length)
        )

    pad_layer = None
    if params:
        weight, bias = extract_params(params)
        kwargs["bias"] = bias is not None
        kwargs["in_channels"] = weight.dims[1] * kwargs.get("groups", 1)
        kwargs["out_channels"] = weight.dims[0]

        if layer_type == "ConvTranspose":
            kwargs["in_channels"], kwargs["out_channels"] = (
                kwargs["out_channels"],
                kwargs["in_channels"],
            )

        # if padding is a layer, remove from kwargs and prepend later
        if "padding" in kwargs and isinstance(kwargs["padding"], nn.Module):
            pad_layer = kwargs.pop("padding")

        # initialize layer and load weights
        layer = layer(**kwargs)
        load_params(layer, weight, bias)
    else:
        # initialize operations without parameters (MaxPool, AvgPool, etc.)
        if layer_type == "MaxPool":
            kwargs["return_indices"] = True

        # if padding is a layer, remove from kwargs and prepend later
        if "padding" in kwargs and isinstance(kwargs["padding"], nn.Module):
            pad_layer = kwargs.pop("padding")
        layer = layer(**kwargs)

    if pad_layer is not None:
        layer = nn.Sequential(pad_layer, layer)

    return layer


def convert_batch_norm_layer(node, params):
    kwargs = extract_attributes(node)
    # Skip input dimension check, not possible before forward pass
    layer = BatchNormWrapper
    torch_params = [torch.from_numpy(numpy_helper.to_array(param)) for param in params]

    # Initialize layer and load weights
    layer = layer(torch_params, **kwargs)
    return layer


def convert_instance_norm_layer(node, params):
    kwargs = extract_attributes(node)
    # Skip input dimension check, not possible before forward pass
    layer = InstanceNormWrapper
    torch_params = [torch.from_numpy(numpy_helper.to_array(param)) for param in params]

    # Initialize layer and load weights
    layer = layer(torch_params, **kwargs)
    return layer


def convert_linear_layer(node, params):
    """Convert linear layer from onnx node and params."""
    # Default Gemm attributes
    dc = dict(
        transpose_weight=True,
        transpose_activation=False,
        weight_multiplier=1,
        bias_multiplier=1,
    )
    dc.update(extract_attributes(node))
    for attr in node.attribute:
        if attr.name in ["transA"] and extract_attr_values(attr) != 0:
            raise NotImplementedError(
                "Not implemented for attr.name={} and value!=0.".format(attr.name)
            )

    kwargs = {}
    weight, bias = extract_params(params)
    kwargs["bias"] = bias is not None
    kwargs["in_features"] = weight.dims[1]
    kwargs["out_features"] = weight.dims[0]

    # initialize layer and load weights
    layer = nn.Linear(**kwargs)
    load_params(layer, weight, bias)

    # apply onnx gemm attributes
    if dc.get("transpose_weight"):
        layer.weight.data = layer.weight.data.t()

    layer.weight.data *= dc.get("weight_multiplier")
    if layer.bias is not None:
        layer.bias.data *= dc.get("bias_multiplier")

    return layer


def extract_and_load_params_lstm(node, weights):
    X = None
    W = None
    R = None
    B = None
    sequence_lens = None
    initial_h = None
    initial_c = None
    P = None

    for par_ix, par_name in enumerate(node.input):
        if par_ix == 0:
            if par_name in weights:
                X = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
        elif par_ix == 1:
            if par_name in weights:
                W = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
        elif par_ix == 2:
            if par_name in weights:
                R = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
        elif par_ix == 3:
            if par_name != "" and par_name in weights:
                B = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
        elif par_ix == 4:
            if par_name != "" and par_name in weights:
                sequence_lens = torch.from_numpy(
                    numpy_helper.to_array(weights[par_name])
                )
        elif par_ix == 5:
            if par_name != "" and par_name in weights:
                initial_h = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
        elif par_ix == 6:
            if par_name != "" and par_name in weights:
                initial_c = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
        elif par_ix == 7:
            if par_name != "" and par_name in weights:
                P = torch.from_numpy(numpy_helper.to_array(weights[par_name]))
    return (X, W, R, B, sequence_lens, initial_h, initial_c, P)


def convert_lstm_layer(node, weights):
    """Convert LSTM layer from onnx node and params."""
    params_tuple = extract_and_load_params_lstm(node, weights)
    (X, W, R, B, sequence_lens, initial_h, initial_c, P) = params_tuple
    if initial_h is not None:
        raise NotImplementedError("LSTM initial_h not yet implemented.")
    if initial_c is not None:
        raise NotImplementedError("LSTM initial_c not yet implemented.")
    if P is not None:
        raise NotImplementedError("LSTM P not yet implemented.")

    dc = dict(
        activation_alpha=None,
        activation_beta=None,
        activations=None,
        clip=None,
        direction="forward",
        hidden_size=None,
        input_forget=0,
        layout=0,
    )
    dc.update(extract_attributes(node))
    if dc["activation_alpha"] is not None:
        raise NotImplementedError(
            "LSTM activation_alpha {}.".format(dc["activation_alpha"])
        )
    if dc["activation_beta"] is not None:
        raise NotImplementedError(
            "LSTM activation_beta {}.".format(dc["activation_beta"])
        )
    if dc["activations"] is not None:
        # TODO allow if torch-compatible activations are set explicitly
        raise NotImplementedError("LSTM activations {}.".format(dc["activations"]))
    if dc["clip"] is not None:
        raise NotImplementedError("LSTM clip {}".format(dc["clip"]))
    if dc["direction"] not in ("forward", "bidirectional"):
        raise ValueError("LSTM direction {}.".format(dc["direction"]))
    if dc["hidden_size"] is None:
        raise ValueError("LSTM hidden_size is None.")
    if dc["input_forget"] != 0:
        raise NotImplementedError("LSTM input_forget {}.".format(dc["input_forget"]))
    if dc["layout"] != 0:
        raise NotImplementedError(
            "LSTM not implemented for layout={}".format(dc["layout"])
        )

    kwargs = {
        "input_size": W.shape[2],
        "hidden_size": dc["hidden_size"],
        "num_layers": 1,
        "bias": True,
        "batch_first": False,
        "dropout": 0,
        "bidirectional": dc["direction"] == "bidirectional",
    }
    lstm_layer = nn.LSTM(**kwargs)

    input_size = kwargs["input_size"]
    hidden_size = kwargs["hidden_size"]
    num_directions = kwargs["bidirectional"] + 1
    num_layers = 1
    if kwargs["bidirectional"]:
        # Set input-hidden weights
        W_iofc = W.transpose(0, 1).view(4 * hidden_size, num_directions, input_size)
        for dir_dim, dir_str in [(0, ""), (1, "_reverse")]:
            W_ifco = torch.cat(
                tensors=(
                    W_iofc[0:hidden_size, dir_dim, :],
                    W_iofc[2 * hidden_size : 4 * hidden_size, dir_dim, :],
                    W_iofc[hidden_size : 2 * hidden_size, dir_dim, :],
                ),
                dim=0,
            )
            getattr(lstm_layer, "weight_ih_l0{}".format(dir_str)).data = W_ifco

        # Set hidden-hidden weights
        R_iofc = R.transpose(0, 1).view(4 * hidden_size, num_directions, hidden_size)
        for dir_dim, dir_str in [(0, ""), (1, "_reverse")]:
            R_ifco = torch.cat(
                tensors=(
                    R_iofc[0:hidden_size, dir_dim, :],
                    R_iofc[2 * hidden_size : 4 * hidden_size, dir_dim, :],
                    R_iofc[hidden_size : 2 * hidden_size, dir_dim, :],
                ),
                dim=0,
            )
            getattr(lstm_layer, "weight_hh_l0{}".format(dir_str)).data = R_ifco

        # Set input-hidden biases
        for dir_dim, dir_str in [(0, ""), (1, "_reverse")]:
            Wb_iofc = B[dir_dim, 0 : 4 * hidden_size]
            Wb_ifco = torch.cat(
                tensors=(
                    Wb_iofc[0:hidden_size],
                    Wb_iofc[2 * hidden_size : 4 * hidden_size],
                    Wb_iofc[hidden_size : 2 * hidden_size],
                ),
                dim=0,
            )
            getattr(lstm_layer, "bias_ih_l0{}".format(dir_str)).data = Wb_ifco

        # Set hidden-hidden biases
        for dir_dim, dir_str in [(0, ""), (1, "_reverse")]:
            Rb_iofc = B[dir_dim, 4 * hidden_size :]
            Rb_ifco = torch.cat(
                tensors=(
                    Rb_iofc[0:hidden_size],
                    Rb_iofc[2 * hidden_size : 4 * hidden_size],
                    Rb_iofc[hidden_size : 2 * hidden_size],
                ),
                dim=0,
            )
            getattr(lstm_layer, "bias_hh_l0{}".format(dir_str)).data = Rb_ifco
    else:
        # Set input-hidden weights
        W_iofc = W.transpose(0, 1).view(4 * hidden_size, input_size)
        W_ifco = torch.cat(
            tensors=(
                W_iofc[0:hidden_size, :],
                W_iofc[2 * hidden_size : 4 * hidden_size, :],
                W_iofc[hidden_size : 2 * hidden_size, :],
            ),
            dim=0,
        )
        getattr(lstm_layer, "weight_ih_l0").data = W_ifco

        # Set hidden-hidden weights
        R_iofc = R.transpose(0, 1).view(4 * hidden_size, hidden_size)
        R_ifco = torch.cat(
            tensors=(
                R_iofc[0:hidden_size, :],
                R_iofc[2 * hidden_size : 4 * hidden_size, :],
                R_iofc[hidden_size : 2 * hidden_size, :],
            ),
            dim=0,
        )
        getattr(lstm_layer, "weight_hh_l0").data = R_ifco

        # Set input-hidden biases
        Wb_iofc = B[0, 0 : 4 * hidden_size]
        Wb_ifco = torch.cat(
            tensors=(
                Wb_iofc[0:hidden_size],
                Wb_iofc[2 * hidden_size : 4 * hidden_size],
                Wb_iofc[hidden_size : 2 * hidden_size],
            ),
            dim=0,
        )
        getattr(lstm_layer, "bias_ih_l0").data = Wb_ifco

        # Set hidden-hidden biases
        Rb_iofc = B[0, 4 * hidden_size :]
        Rb_ifco = torch.cat(
            tensors=(
                Rb_iofc[0:hidden_size],
                Rb_iofc[2 * hidden_size : 4 * hidden_size],
                Rb_iofc[hidden_size : 2 * hidden_size],
            ),
            dim=0,
        )
        getattr(lstm_layer, "bias_hh_l0").data = Rb_ifco

    layer = LSTMWrapper(lstm_layer)
    return layer



from collections import defaultdict
from copy import deepcopy
from functools import partial
from importlib import import_module
import warnings

import numpy as np
import onnx
import torch
from onnx import numpy_helper
from torch import nn
from torch.nn.modules.linear import Identity

from onnx2pytorch.utils import (
    get_inputs_names,
    get_outputs_names,
)


class Loop(nn.Module):
    def __init__(
        self,
        opset_version,
        batch_dim,
        body: onnx.GraphProto,
    ):
        super().__init__()
        self.ops = import_module("onnx2pytorch.convert.operations")
        self.c = import_module("onnx2pytorch.constants")

        self.body = body
        self.batch_dim = batch_dim

        self.input_names = get_inputs_names(body)
        self.output_names = get_outputs_names(body)

        # Creates mapping from node (identified by first output) to submodule
        self.mapping = {}
        for op_id, op_name, op in self.ops.convert_operations(
            body, opset_version, batch_dim
        ):
            setattr(self, op_name, op)
            self.mapping[op_id] = op_name

        # Store initializers as buffers
        for tensor in self.body.initializer:
            buffer_name = self.ops.get_buffer_name(tensor.name)
            self.register_buffer(
                buffer_name,
                torch.from_numpy(numpy_helper.to_array(tensor)),
            )

        # We do not track dependencies (for memory reduction) within loops.
        # This would be complicated due to loop-carried dependencies.

    def forward(self, enclosing_modules, enclosing_activations, *inputs):
        """
        Parameters
        ----------
        enclosing_modules: tuple of nn.Modules
            Module(s) from enclosing scope(s), containing initializers as buffers.
        enclosing_activations: dict
            All activations from the enclosing scope.
        inputs: list
            Inputs to Loop node (length >= 2), comprising M, cond, and v_initial.

        Returns
        -------
        v_final_and_scan_outputs: list
            Final N loop carried dependency values, then K scan_outputs.
        """

        N = len(self.input_names) - 2
        K = len(self.output_names) - (1 + N)

        M = inputs[0]
        cond = inputs[1]
        v_initial = inputs[2:]

        iteration_num_name = self.input_names[0]
        cond_in_name = self.input_names[1]
        loop_carried_in_names = self.input_names[2:]
        cond_out_name = self.output_names[0]
        loop_carried_out_names = self.output_names[1 : N + 1]
        scan_outputs_names = self.output_names[1 + N :]

        buffer_modules = enclosing_modules + (self,)

        activations = {}
        activations.update(zip(loop_carried_in_names, v_initial))
        activations.update(enclosing_activations)

        scan_outputs = defaultdict(list)
        i = torch.tensor(0)
        while i < M and cond:
            activations[iteration_num_name] = i
            activations[cond_in_name] = cond
            for node in self.body.node:
                # Identifying the layer ids and names
                out_op_id = node.output[0]
                out_op_name = self.mapping[out_op_id]
                in_op_names = [
                    self.mapping.get(in_op_id, in_op_id)
                    for in_op_id in node.input
                    if in_op_id in activations
                ]

                # getting correct layer
                op = getattr(self, out_op_name)

                # if first layer choose input as in_activations
                # if not in_op_names and len(node.input) == 1:
                #    in_activations = input
                if isinstance(op, self.c.STANDARD_LAYERS) or (
                    isinstance(op, self.c.COMPOSITE_LAYERS)
                    and any(isinstance(x, self.c.STANDARD_LAYERS) for x in op.modules())
                ):
                    in_activations = [
                        activations[in_op_id]
                        for in_op_id in node.input
                        if in_op_id in activations
                    ]
                else:
                    in_activations = [
                        activations[in_op_id] if in_op_id in activations
                        # if in_op_id not in activations neither in parameters then
                        # it must be the initial input
                        else self.ops.get_init_parameter(
                            buffer_modules, in_op_id, inputs[0]
                        )
                        for in_op_id in node.input
                    ]

                in_activations = [
                    in_act for in_act in in_activations if in_act is not None
                ]

                # store activations for next layer
                if isinstance(op, Loop):
                    outputs = op(buffer_modules, activations, *in_activations)
                    for out_act_name, output in zip(node.output, outputs):
                        activations[out_op_id] = output
                        if out_act_name in scan_outputs_names:
                            scan_outputs[out_act_name].append(output)
                elif isinstance(op, partial) and op.func == torch.cat:
                    output = op(in_activations)
                    activations[out_op_id] = output
                    if out_op_id in scan_outputs_names:
                        scan_outputs[out_op_id].append(output)
                elif isinstance(op, Identity):
                    # After batch norm fusion the batch norm parameters
                    # were all passed to identity instead of first one only
                    output = op(in_activations[0])
                    activations[out_op_id] = op(output)
                    if out_op_id in scan_outputs_names:
                        scan_outputs[out_op_id].append(output)
                elif isinstance(op, self.c.MULTIOUTPUT_LAYERS) or (
                    isinstance(op, self.c.COMPOSITE_LAYERS)
                    and any(
                        isinstance(x, self.c.MULTIOUTPUT_LAYERS) for x in op.modules()
                    )
                ):
                    outputs = op(*in_activations)
                    for out_act_name, output in zip(node.output, outputs):
                        activations[out_act_name] = output
                        if out_act_name in scan_outputs_names:
                            scan_outputs[out_act_name].append(output)
                else:
                    output = op(*in_activations)
                    activations[out_op_id] = output
                    if out_op_id in scan_outputs_names:
                        scan_outputs[out_op_id].append(output)

            # Prepare for next iteration
            cond = activations[cond_out_name]
            i += 1
            loop_carried_outputs = [
                activations[aname] for aname in loop_carried_out_names
            ]
            activations.update(zip(loop_carried_in_names, loop_carried_outputs))

        # Set outputs to N loop carried final values, followed by K scan outputs
        outputs = [activations[lcn] for lcn in loop_carried_out_names]
        for son in scan_outputs_names:
            outputs.append(torch.cat([so.unsqueeze(dim=0) for so in scan_outputs[son]]))
        return outputs



from torch import nn


class LSTMWrapper(nn.Module):
    """Wraps a 1-layer nn.LSTM to match the API of an ONNX LSTM.

    It expects h_0 and c_0 as separate inputs rather than as a tuple,
    and returns h_n and c_n as separate outputs rather than as a tuple.
    """

    def __init__(self, lstm_module: nn.LSTM):
        super().__init__()
        self.lstm = lstm_module

    def forward(self, input, h_0=None, c_0=None):
        (seq_len, batch, input_size) = input.shape
        num_layers = 1
        num_directions = self.lstm.bidirectional + 1
        hidden_size = self.lstm.hidden_size
        if h_0 is None or c_0 is None or h_0.numel() == 0 or c_0.numel() == 0:
            tuple_0 = None
        else:
            tuple_0 = (h_0, c_0)
        output, (h_n, c_n) = self.lstm(input, tuple_0)

        # Y has shape (seq_length, num_directions, batch_size, hidden_size)
        Y = output.view(seq_len, batch, num_directions, hidden_size).transpose(1, 2)
        # Y_h has shape (num_directions, batch_size, hidden_size)
        Y_h = h_n.view(num_layers, num_directions, batch, hidden_size).squeeze(0)
        # Y_c has shape (num_directions, batch_size, hidden_size)
        Y_c = c_n.view(num_layers, num_directions, batch, hidden_size).squeeze(0)
        return Y, Y_h, Y_c



import torch
from torch import nn


class Gather(nn.Module):
    def __init__(self, dim=0):
        self.dim = dim
        self.selection = [slice(None) for _ in range(dim)]
        super().__init__()

    def forward(self, data: torch.Tensor, indices: torch.Tensor):
        selection = self.selection + [indices.to(torch.int64)]
        return data.__getitem__(selection)



import torch
from torch import nn


class Div(nn.Module):
    def forward(self, input, other):
        res_type = torch.result_type(input, other)
        true_quotient = torch.true_divide(input, other)
        if res_type.is_floating_point:
            res = true_quotient
        else:
            res = torch.floor(true_quotient).to(res_type)
        return res



import warnings

import onnx
from onnx import numpy_helper

from onnx2pytorch.utils import (
    extract_padding_params_for_conv_layer,
    extract_padding_params,
)

TENSOR_PROTO_MAPPING = dict([i[::-1] for i in onnx.TensorProto.DataType.items()])

AttributeType = dict(
    UNDEFINED=0,
    FLOAT=1,
    INT=2,
    STRING=3,
    TENSOR=4,
    GRAPH=5,
    SPARSE_TENSOR=11,
    FLOATS=6,
    INTS=7,
    STRINGS=8,
    TENSORS=9,
    GRAPHS=10,
    SPARSE_TENSORS=12,
)


def extract_attr_values(attr):
    """Extract onnx attribute values."""
    if attr.type == AttributeType["INT"]:
        value = attr.i
    elif attr.type == AttributeType["FLOAT"]:
        value = attr.f
    elif attr.type == AttributeType["INTS"]:
        value = tuple(attr.ints)
    elif attr.type == AttributeType["FLOATS"]:
        value = tuple(attr.floats)
    elif attr.type == AttributeType["TENSOR"]:
        value = numpy_helper.to_array(attr.t)
    elif attr.type == AttributeType["STRING"]:
        value = attr.s.decode()
    elif attr.type == AttributeType["GRAPH"]:
        value = attr.g
    else:
        raise NotImplementedError(
            "Extraction of attribute type {} not implemented.".format(attr.type)
        )
    return value


def extract_attributes(node):
    """Extract onnx attributes. Map onnx feature naming to pytorch."""
    kwargs = {}
    for attr in node.attribute:
        if attr.name == "activation_alpha":
            kwargs["activation_alpha"] = extract_attr_values(attr)
        elif attr.name == "activation_beta":
            kwargs["activation_beta"] = extract_attr_values(attr)
        elif attr.name == "activations":
            kwargs["activations"] = extract_attr_values(attr)
        elif attr.name == "alpha":
            if node.op_type == "LeakyRelu":
                kwargs["negative_slope"] = extract_attr_values(attr)
            elif node.op_type in ("Elu", "ThresholdedRelu"):
                kwargs["alpha"] = extract_attr_values(attr)
            else:
                kwargs["weight_multiplier"] = extract_attr_values(attr)
        elif attr.name == "auto_pad":
            value = extract_attr_values(attr)
            if value == "NOTSET":
                pass
            else:
                raise NotImplementedError(
                    "auto_pad={} functionality not implemented.".format(value)
                )
        elif attr.name == "axis" and node.op_type == "Flatten":
            kwargs["start_dim"] = extract_attr_values(attr)
        elif attr.name == "axis" or attr.name == "axes":
            v = extract_attr_values(attr)
            if isinstance(v, (tuple, list)) and len(v) == 1:
                kwargs["dim"] = v[0]
            else:
                kwargs["dim"] = v
        elif attr.name == "beta":
            kwargs["bias_multiplier"] = extract_attr_values(attr)
        elif attr.name == "body":
            kwargs["body"] = extract_attr_values(attr)
        elif attr.name == "ceil_mode":
            kwargs["ceil_mode"] = bool(extract_attr_values(attr))
        elif attr.name == "center_point_box":
            kwargs["center_point_box"] = extract_attr_values(attr)
        elif attr.name == "clip":
            kwargs["clip"] = extract_attr_values(attr)
        elif attr.name == "coordinate_transformation_mode":
            arg = extract_attr_values(attr)
            if arg == "align_corners":
                kwargs["align_corners"] = True
            else:
                warnings.warn(
                    "Pytorch's interpolate uses no coordinate_transformation_mode={}. "
                    "Result might differ.".format(arg)
                )
        elif attr.name == "dilations":
            kwargs["dilation"] = extract_attr_values(attr)
        elif attr.name == "direction":
            kwargs["direction"] = extract_attr_values(attr)
        elif attr.name == "ends":
            kwargs["ends"] = extract_attr_values(attr)
        elif attr.name == "epsilon":
            kwargs["eps"] = extract_attr_values(attr)
        elif attr.name == "group":
            kwargs["groups"] = extract_attr_values(attr)
        elif attr.name == "hidden_size":
            kwargs["hidden_size"] = extract_attr_values(attr)
        elif attr.name == "input_forget":
            kwargs["input_forget"] = extract_attr_values(attr)
        elif attr.name == "keepdims":
            kwargs["keepdim"] = bool(extract_attr_values(attr))
        elif attr.name == "kernel_shape":
            kwargs["kernel_size"] = extract_attr_values(attr)
        elif attr.name == "largest":
            kwargs["largest"] = extract_attr_values(attr)
        elif attr.name == "layout":
            kwargs["layout"] = extract_attr_values(attr)
        elif attr.name == "max":
            kwargs["max"] = extract_attr_values(attr)
        elif attr.name == "min":
            kwargs["min"] = extract_attr_values(attr)
        elif attr.name == "mode":
            kwargs["mode"] = extract_attr_values(attr)
        elif attr.name == "momentum":
            kwargs["momentum"] = extract_attr_values(attr)
        elif attr.name == "noop_with_empty_axes":
            kwargs["noop_with_empty_axes"] = extract_attr_values(attr)
        elif attr.name == "output_shape" and node.op_type == "ConvTranspose":
            raise NotImplementedError(
                "ConvTranspose with dynamic padding not implemented."
            )
        elif attr.name == "pads":
            params = extract_attr_values(attr)
            if node.op_type == "Pad":
                kwargs["padding"] = extract_padding_params(params)
            else:
                # Works for Conv, MaxPooling and other layers from convert_layer func
                kwargs["padding"] = extract_padding_params_for_conv_layer(params)
        elif attr.name == "perm":
            kwargs["dims"] = extract_attr_values(attr)
        elif attr.name == "repeats":
            kwargs["repeats"] = extract_attr_values(attr)
        elif attr.name == "sorted":
            kwargs["sorted"] = extract_attr_values(attr)
        elif attr.name == "sparse_value":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "spatial":
            kwargs["spatial"] = extract_attr_values(attr)  # Batch norm parameter
        elif attr.name == "split":
            kwargs["split_size_or_sections"] = extract_attr_values(attr)
        elif attr.name == "strides":
            kwargs["stride"] = extract_attr_values(attr)
        elif attr.name == "starts":
            kwargs["starts"] = extract_attr_values(attr)
        elif attr.name == "to":
            kwargs["dtype"] = TENSOR_PROTO_MAPPING[extract_attr_values(attr)].lower()
        elif attr.name == "transB":
            kwargs["transpose_weight"] = not extract_attr_values(attr)
        elif attr.name == "transA":
            kwargs["transpose_activation"] = bool(extract_attr_values(attr))
        elif attr.name == "value":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "value_float":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "value_floats":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "value_int":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "value_ints":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "value_string":
            kwargs["constant"] = extract_attr_values(attr)
        elif attr.name == "value_strings":
            kwargs["constant"] = extract_attr_values(attr)
        elif node.op_type == "Resize":
            # These parameters are not used, warn in Resize operator
            kwargs[attr.name] = extract_attr_values(attr)
        else:
            raise NotImplementedError(
                "Extraction of attribute {} not implemented.".format(attr.name)
            )
    return kwargs



from collections import defaultdict
from functools import partial

import numpy as np
import onnx
import torch
from torch import nn
from torch.nn import functional as F
from onnx import numpy_helper
from torch.nn.modules.linear import Identity

from onnx2pytorch.convert.attribute import extract_attributes
from onnx2pytorch.convert.layer import (
    convert_layer,
    convert_linear_layer,
    convert_batch_norm_layer,
    convert_instance_norm_layer,
    convert_lstm_layer,
)
from onnx2pytorch.operations import *
from onnx2pytorch.operations.base import OperatorWrapper
from onnx2pytorch.operations import Resize, Upsample
from onnx2pytorch.utils import (
    get_inputs_names,
    get_outputs_names,
    value_wrapper,
)


def get_buffer_name(param_name):
    """
    Convert name of initializer to valid name for nn.Module attribute.
    """
    return "_initializer_{}".format(param_name.replace(".", "_"))


def get_init_parameter(modules, item, default):
    """
    Look in modules for the item, and if not found return default.

    Parameters
    ----------
    modules: list of nn.Modules
        Modules whose attributes to search.
    item: str
        Name of initializer in ONNX model.
    default: torch.Tensor
        Tensor to return if item not found.
    """
    item_name = get_buffer_name(item)
    for mod in modules:
        if hasattr(mod, item_name):
            return getattr(mod, item_name)
    return default


def convert_operations(onnx_graph, opset_version, batch_dim=0, enable_pruning=True):
    """
    Convert onnx model operations. Yields onnx's operator_id, operator_name and
    converted pytorch operator.

    Parameters
    ----------
    onnx_graph: onnx.GraphProto
        Loaded onnx model's GraphProto.
    opset_version: int
        ONNX model's opset version.
    batch_dim: int
        Usually 0 for computer vision models and 1 for NLP models.
    enable_pruning: bool
        Track kept/pruned indices between different calls to forward pass.

    Returns
    -------
    iterator: (op_id, op_name, op)
    """
    weights = {tensor.name: tensor for tensor in onnx_graph.initializer}

    for i, node in enumerate(onnx_graph.node):
        # extract only useful inputs
        params = [weights[par_name] for par_name in node.input if par_name in weights]

        if node.op_type == "Add":
            op = Add(feature_dim=batch_dim + 1)  # 0 for CV models and 1 for NLP
        elif node.op_type == "And":
            op = OperatorWrapper(torch.logical_and)
        elif node.op_type == "AveragePool":
            op = convert_layer(node, "AvgPool")
        elif node.op_type == "BatchNormalization":
            op = convert_batch_norm_layer(node, params=params)
        elif node.op_type == "Cast":
            op = Cast(**extract_attributes(node))
        elif node.op_type == "Ceil":
            op = OperatorWrapper(torch.ceil)
        elif node.op_type == "Clip":
            op = Clip(**extract_attributes(node))
        elif node.op_type == "Concat":
            op = partial(torch.cat, **extract_attributes(node))
        elif node.op_type == "Constant":
            op = Constant(**extract_attributes(node))
        elif node.op_type == "ConstantOfShape":
            op = ConstantOfShape(**extract_attributes(node))
        elif node.op_type == "Conv":
            op = convert_layer(node, "Conv", params)
        elif node.op_type == "ConvTranspose":
            op = convert_layer(node, "ConvTranspose", params)
        elif node.op_type == "Div":
            op = Div()
        elif node.op_type == "Elu":
            op = nn.ELU(**extract_attributes(node), inplace=True)
        elif node.op_type == "Equal":
            op = OperatorWrapper(torch.eq)
        elif node.op_type == "Erf":
            op = OperatorWrapper(torch.erf)
        elif node.op_type == "Exp":
            op = OperatorWrapper(torch.exp)
        elif node.op_type == "Expand":
            op = Expand()
        elif node.op_type == "Flatten":
            op = Flatten(**extract_attributes(node))
            op.feature_dim = batch_dim + 1  # Necessary for transformers
        elif node.op_type == "Floor":
            op = OperatorWrapper(torch.floor)
        elif node.op_type == "Gather":
            op = Gather(**extract_attributes(node))
        elif node.op_type == "GatherND":
            op = GatherND(**extract_attributes(node))
        elif node.op_type == "Gemm":
            op = convert_linear_layer(node, params)
        elif node.op_type == "GlobalAveragePool":
            op = GlobalAveragePool()
        elif node.op_type == "Greater":
            op = OperatorWrapper(torch.greater)
        elif node.op_type == "GreaterOrEqual":
            op = OperatorWrapper(torch.greater_equal)
        elif node.op_type == "Identity":
            op = nn.Identity()
        elif node.op_type == "InstanceNormalization":
            op = convert_instance_norm_layer(node, params=params)
        elif node.op_type == "LeakyRelu":
            op = nn.LeakyReLU(**extract_attributes(node), inplace=True)
        elif node.op_type == "Less":
            op = OperatorWrapper(torch.less)
        elif node.op_type == "Log":
            op = OperatorWrapper(torch.log)
        elif node.op_type == "Loop":
            op = Loop(
                opset_version=opset_version,
                batch_dim=batch_dim,
                **extract_attributes(node),
            )
        elif node.op_type == "LSTM":
            op = convert_lstm_layer(node, weights)
        elif node.op_type == "MatMul":
            if params:
                weight = torch.from_numpy(numpy_helper.to_array(params[0]))
                op = nn.Linear(weight.shape[0], weight.shape[1], bias=False)
                op.weight.data = weight.t()

                # check if next node Add to add bias
                next_node = onnx_graph.node[i + 1]
                next_params = [
                    weights[par_name]
                    for par_name in next_node.input
                    if par_name in weights
                ]
                if next_params and next_node.op_type == "Add":
                    bias = torch.from_numpy(numpy_helper.to_array(next_params[0]))
                    op.bias = nn.Parameter(bias)
                    node.output.pop()
                    node.output.extend(next_node.output)
                    onnx_graph.node.pop(i + 1)  # remove next node
            else:
                op = MatMul()
        elif node.op_type == "Max":
            op = OperatorWrapper(torch.max)
        elif node.op_type == "MaxPool":
            op = convert_layer(node, "MaxPool")
        elif node.op_type == "Min":
            op = OperatorWrapper(torch.min)
        elif node.op_type == "Mul":
            op = OperatorWrapper(torch.mul)
        elif node.op_type == "NonMaxSuppression":
            op = NonMaxSuppression(**extract_attributes(node))
        elif node.op_type == "Not":
            op = OperatorWrapper(torch.logical_not)
        elif node.op_type == "OneHot":
            op = OneHot(**extract_attributes(node))
        elif node.op_type == "Or":
            op = OperatorWrapper(torch.logical_or)
        elif node.op_type == "Pad":
            op = Pad(**extract_attributes(node))
        elif node.op_type == "Pow":
            op = OperatorWrapper(torch.pow)
        elif node.op_type == "PRelu":
            op = PRelu()
        elif node.op_type == "Range":
            op = Range()
        elif node.op_type == "Reciprocal":
            op = OperatorWrapper(torch.reciprocal)
        elif node.op_type == "ReduceMax":
            op = ReduceMax(**extract_attributes(node))
        elif node.op_type == "ReduceMean":
            kwargs = dict(keepdim=True)
            kwargs.update(extract_attributes(node))
            op = partial(torch.mean, **kwargs)
        elif node.op_type == "ReduceMin":
            kwargs = dict(keepdim=True)
            kwargs.update(extract_attributes(node))
            op = partial(torch.min, **kwargs)
        elif node.op_type == "ReduceProd":
            kwargs = dict(keepdim=True)
            kwargs.update(extract_attributes(node))
            op = partial(torch.prod, **kwargs)
        elif node.op_type == "ReduceSum":
            op = ReduceSum(opset_version=opset_version, **extract_attributes(node))
        elif node.op_type == "ReduceL2":
            op = ReduceL2(opset_version=opset_version, **extract_attributes(node))
        elif node.op_type == "Relu":
            op = nn.ReLU(inplace=True)
        elif node.op_type == "Reshape":
            shape = list(
                filter(lambda x: x.name == node.input[1], onnx_graph.initializer)
            )
            shape = np.copy(numpy_helper.to_array(shape[0])) if shape else None
            op = Reshape(enable_pruning, shape)
        elif node.op_type == "Resize":
            op = Resize(**extract_attributes(node))
        elif node.op_type == "Scatter":
            op = Scatter(**extract_attributes(node))
        elif node.op_type == "ScatterElements":
            op = ScatterElements(**extract_attributes(node))
        elif node.op_type == "ScatterND":
            op = ScatterND()
        elif node.op_type == "Shape":
            op = Shape()
        elif node.op_type == "Sigmoid":
            op = nn.Sigmoid()
        elif node.op_type == "Slice":
            op = Slice(**extract_attributes(node))
        elif node.op_type == "Softmax":
            kwargs = dict(dim=-1)
            kwargs.update(extract_attributes(node))
            op = nn.Softmax(**kwargs)
        elif node.op_type == "Softplus":
            op = nn.Softplus(beta=1)
        elif node.op_type == "Softsign":
            op = nn.Softsign()
        elif node.op_type == "Split":
            kwargs = extract_attributes(node)
            # if the split_size_or_sections is not in node attributes,
            # the number_of_splits becomes the number of node outputs
            if "split_size_or_sections" not in kwargs:
                kwargs["number_of_splits"] = len(node.output)
            op = Split(enable_pruning, **kwargs)
        elif node.op_type == "Sqrt":
            op = OperatorWrapper(torch.sqrt)
        elif node.op_type == "Squeeze":
            op = Squeeze(opset_version=opset_version, **extract_attributes(node))
        elif node.op_type == "Sub":
            op = OperatorWrapper(torch.sub)
        elif node.op_type == "Tanh":
            op = OperatorWrapper(torch.tanh)
        elif node.op_type == "ThresholdedRelu":
            op = ThresholdedRelu(**extract_attributes(node))
        elif node.op_type == "Tile":
            op = Tile()
        elif node.op_type == "TopK":
            op = TopK()
        elif node.op_type == "Transpose":
            op = Transpose(**extract_attributes(node))
        elif node.op_type == "Unsqueeze":
            op = Unsqueeze(opset_version=opset_version, **extract_attributes(node))
        elif node.op_type == "Upsample":
            op = Upsample(**extract_attributes(node))
        elif node.op_type == "Where":
            op = Where()
        else:
            op = getattr(torch, node.op_type.lower(), None)
            if op is None:
                raise NotImplementedError(
                    "Conversion not implemented for op_type={}.".format(node.op_type)
                )
            else:
                print(
                    "Automatic inference of operator: {}".format(node.op_type.lower())
                )

        op_name = "{}_{}".format(node.op_type, node.output[0])
        op_id = node.output[0]
        yield op_id, op_name, op



import torch
from torch import nn


class Range(nn.Module):
    def forward(self, start: torch.Tensor, limit: torch.Tensor, delta: torch.Tensor):
        return torch.arange(start=start, end=limit, step=delta)



from collections import defaultdict
from copy import deepcopy
from functools import partial
import warnings

import numpy as np
import onnx
import torch
from onnx import numpy_helper
from torch import nn
from torch.jit import TracerWarning
from torch.nn.modules.linear import Identity

from onnx2pytorch.constants import (
    COMPOSITE_LAYERS,
    MULTIOUTPUT_LAYERS,
    STANDARD_LAYERS,
)
from onnx2pytorch.convert.debug import debug_model_conversion
from onnx2pytorch.convert.operations import (
    convert_operations,
    get_buffer_name,
    get_init_parameter,
    Loop,
)
from onnx2pytorch.utils import (
    get_inputs_names,
    get_outputs_names,
)


def compute_activation_dependencies(onnx_graph, model, mapping):
    """
    Compute activation dependencies, mapping each node to its dependents.

    Parameters
    ----------
    onnx_graph: onnx.GraphProto
        ONNX graph.
    model: onnx2pytorch.ConvertModel
        Module which contains converted submodules.
    mapping: dict
        Dictionary mapping from node name to name of submodule.

    Returns
    -------
    needed_by: dict
        Dictionary mapping from node name to names of its dependents.
    """
    needed_by = defaultdict(set)
    for node in onnx_graph.node:
        out_op_id = node.output[0]
        for in_op_id in node.input:
            needed_by[in_op_id].add(out_op_id)
        if node.op_type == "Loop":
            # Look at nodes in the loop body
            l1 = getattr(model, mapping[out_op_id])  # Loop object
            loop_body_l1 = l1.body
            for node_l1 in loop_body_l1.node:
                for in_op_id in node_l1.input:
                    # Treating node (outer loop) as dependent, not node_l1
                    needed_by[in_op_id].add(out_op_id)
                if node_l1.op_type == "Loop":
                    # Look at nodes in the loop body
                    l2 = getattr(model, l1.mapping[node_l1.output[0]])  # Loop object
                    loop_body_l2 = l2.body
                    for node_l2 in loop_body_l2.node:
                        for in_op_id in node_l2.input:
                            # Treating node (outer loop) as dependent, not node_l2
                            needed_by[in_op_id].add(out_op_id)
                        if node_l2.op_type == "Loop":
                            # TODO: make this recursive for nested loops
                            raise NotImplementedError(
                                "Activation garbage collection not implemented for >2 nested loops."
                            )
    needed_by.default_factory = None
    return needed_by


class ConvertModel(nn.Module):
    def __init__(
        self,
        onnx_model: onnx.ModelProto,
        batch_dim=0,
        experimental=False,
        debug=False,
        enable_pruning=False,
    ):
        """
        Convert onnx model to pytorch.

        Parameters
        ----------
        onnx_model: onnx.ModelProto
            Loaded onnx model.
        batch_dim: int
            Dimension of the batch.
        experimental: bool
            Experimental implementation allows batch_size > 1. However,
            batchnorm layers could potentially produce false outputs.
        enable_pruning: bool
            Track kept/pruned indices between different calls to forward pass.

        Returns
        -------
        model: torch.nn.Module
            A converted pytorch model.
        """
        super().__init__()
        self.onnx_model = onnx_model
        self.batch_dim = batch_dim
        self.experimental = experimental
        self.debug = debug
        self.enable_pruning = enable_pruning

        self.input_names = get_inputs_names(onnx_model.graph)
        self.output_names = get_outputs_names(onnx_model.graph)
        opset_version = onnx_model.opset_import[0].version

        # Create mapping from node (identified by first output) to submodule
        self.mapping = {}
        for op_id, op_name, op in convert_operations(
            onnx_model.graph,
            opset_version,
            batch_dim,
            enable_pruning,
        ):
            setattr(self, op_name, op)
            if isinstance(op, Loop) and debug:
                raise NotImplementedError("debug-mode with Loop node not implemented.")
            self.mapping[op_id] = op_name

        # Store initializers as buffers
        for tensor in self.onnx_model.graph.initializer:
            buffer_name = get_buffer_name(tensor.name)
            self.register_buffer(
                buffer_name,
                torch.from_numpy(numpy_helper.to_array(tensor)),
            )

        # Compute activation dependencies, mapping each node to its dependents
        self.needed_by = compute_activation_dependencies(
            self.onnx_model.graph, self, self.mapping
        )

        if experimental:
            warnings.warn(
                "Using experimental implementation that allows 'batch_size > 1'."
                "Batchnorm layers could potentially produce false outputs."
            )

    def forward(self, *input_list, **input_dict):
        if len(input_list) > 0 and len(input_dict) > 0:
            raise ValueError(
                "forward-pass accepts either input_list (positional args) or "
                "input_dict (keyword args) but not both"
            )
        if len(input_list) > 0:
            inputs = input_list
        if len(input_dict) > 0:
            inputs = [input_dict[key] for key in self.input_names]

        if not self.experimental and inputs[0].shape[self.batch_dim] > 1:
            raise NotImplementedError(
                "Input with larger batch size than 1 not supported yet."
            )
        activations = dict(zip(self.input_names, inputs))
        still_needed_by = deepcopy(self.needed_by)

        for node in self.onnx_model.graph.node:
            # Identifying the layer ids and names
            out_op_id = node.output[0]
            out_op_name = self.mapping[out_op_id]
            in_op_names = [
                self.mapping.get(in_op_id, in_op_id)
                for in_op_id in node.input
                if in_op_id in activations
            ]

            # getting correct layer
            op = getattr(self, out_op_name)

            # if first layer choose input as in_activations
            # if not in_op_names and len(node.input) == 1:
            #    in_activations = input
            if isinstance(op, STANDARD_LAYERS) or (
                isinstance(op, COMPOSITE_LAYERS)
                and any(isinstance(x, STANDARD_LAYERS) for x in op.modules())
            ):
                in_activations = [
                    activations[in_op_id]
                    for in_op_id in node.input
                    if in_op_id in activations
                ]
            else:
                in_activations = [
                    activations[in_op_id] if in_op_id in activations
                    # if in_op_id not in activations neither in parameters then
                    # it must be the initial input
                    else get_init_parameter([self], in_op_id, inputs[0])
                    for in_op_id in node.input
                ]

            in_activations = [in_act for in_act in in_activations if in_act is not None]

            # store activations for next layer
            if isinstance(op, Loop):
                outputs = op((self,), activations, *in_activations)
                for out_op_id, output in zip(node.output, outputs):
                    activations[out_op_id] = output
            elif isinstance(op, partial) and op.func == torch.cat:
                activations[out_op_id] = op(in_activations)
            elif isinstance(op, Identity):
                # After batch norm fusion the batch norm parameters
                # were all passed to identity instead of first one only
                activations[out_op_id] = op(in_activations[0])
            elif isinstance(op, MULTIOUTPUT_LAYERS) or (
                isinstance(op, COMPOSITE_LAYERS)
                and any(isinstance(x, MULTIOUTPUT_LAYERS) for x in op.modules())
            ):
                for out_op_id, output in zip(node.output, op(*in_activations)):
                    activations[out_op_id] = output
            else:
                activations[out_op_id] = op(*in_activations)

            # Remove activations that are no longer needed
            for in_op_id in node.input:
                if in_op_id in still_needed_by:
                    still_needed_by[in_op_id].discard(out_op_id)
                    if len(still_needed_by[in_op_id]) == 0:
                        if in_op_id in activations:
                            del activations[in_op_id]

            if self.debug:
                # compare if the activations of pytorch are the same as from onnxruntime
                debug_model_conversion(
                    self.onnx_model,
                    [activations[x] for x in self.input_names],
                    [activations[out_op_id] for out_op_id in node.output],
                    node,
                )

        # collect all outputs
        outputs = [activations[x] for x in self.output_names]
        if len(outputs) == 1:
            outputs = outputs[0]
        return outputs



import numpy as np
import torch
from torch import nn


class Constant(nn.Module):
    def __init__(self, constant):
        super().__init__()
        self.register_buffer("constant", torch.from_numpy(np.copy(constant)))

    def forward(self):
        return self.constant

    def extra_repr(self) -> str:
        return "constant={}".format(self.constant)



import torch
from torch import nn


class GlobalAveragePool(nn.Module):
    def forward(self, input: torch.Tensor):
        spatial_shape = input.ndimension() - 2
        dim = tuple(range(2, spatial_shape + 2))
        return torch.mean(input, dim=dim, keepdim=True)



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

try:
    with open("README.md", "r", encoding="utf-8") as f:
        long_description = f.read()
except IOError:
    long_description = ""

# Extract version. Cannot import directly because of import error.
root_dir = os.path.dirname(__file__)
with open(os.path.join(root_dir, "onnx2pytorch/__init__.py"), "r") as f:
    for line in f.readlines():
        if line.startswith("__version__"):
            version = line.split("=")[-1].strip().strip('"')
            break

PACKAGES = find_packages(exclude=("tests.*", "tests"))

setup(
    name="onnx2pytorch",
    version=version,
    description="Library to transform onnx model to pytorch.",
    license="apache-2.0",
    author="Talmaj Marinc",
    packages=PACKAGES,
    install_requires=["torch>=1.4.0", "onnx>=1.6.0", "torchvision>=0.9.0"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ToriML/onnx2pytorch",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import numpy as np
import torch
from torch import nn


class ConstantOfShape(nn.Module):
    def __init__(self, constant=None):
        super().__init__()
        if constant is None:
            const = torch.tensor(1.0, dtype=torch.float32)
        else:
            const = torch.from_numpy(np.copy(constant))
        self.register_buffer("constant", const)

    def forward(self, shape: torch.Tensor):
        return self.constant.expand(*shape).to(shape.device)

    def extra_repr(self) -> str:
        return "constant={}".format(self.constant)



import warnings
import torch

from torch.nn.modules.instancenorm import _InstanceNorm

try:
    from torch.nn.modules.batchnorm import _LazyNormBase

    class _LazyInstanceNorm(_LazyNormBase, _InstanceNorm):
        cls_to_become = _InstanceNorm

except ImportError:
    from torch.nn.modules.lazy import LazyModuleMixin
    from torch.nn.parameter import UninitializedBuffer, UninitializedParameter

    class _LazyInstanceNorm(LazyModuleMixin, _InstanceNorm):
        weight: UninitializedParameter  # type: ignore[assignment]
        bias: UninitializedParameter  # type: ignore[assignment]

        cls_to_become = _InstanceNorm

        def __init__(
            self,
            eps=1e-5,
            momentum=0.1,
            affine=True,
            track_running_stats=True,
            device=None,
            dtype=None,
        ) -> None:
            factory_kwargs = {"device": device, "dtype": dtype}
            super(_LazyInstanceNorm, self).__init__(
                # affine and track_running_stats are hardcoded to False to
                # avoid creating tensors that will soon be overwritten.
                0,
                eps,
                momentum,
                False,
                False,
                **factory_kwargs,
            )
            self.affine = affine
            self.track_running_stats = track_running_stats
            if self.affine:
                self.weight = UninitializedParameter(**factory_kwargs)
                self.bias = UninitializedParameter(**factory_kwargs)
            if self.track_running_stats:
                self.running_mean = UninitializedBuffer(**factory_kwargs)
                self.running_var = UninitializedBuffer(**factory_kwargs)
                self.num_batches_tracked = torch.tensor(
                    0,
                    dtype=torch.long,
                    **{k: v for k, v in factory_kwargs.items() if k != "dtype"},
                )

        def reset_parameters(self) -> None:
            if not self.has_uninitialized_params() and self.num_features != 0:
                super().reset_parameters()

        def initialize_parameters(self, input) -> None:  # type: ignore[override]
            if self.has_uninitialized_params():
                self.num_features = input.shape[1]
                if self.affine:
                    assert isinstance(self.weight, UninitializedParameter)
                    assert isinstance(self.bias, UninitializedParameter)
                    self.weight.materialize((self.num_features,))
                    self.bias.materialize((self.num_features,))
                if self.track_running_stats:
                    self.running_mean.materialize(
                        (self.num_features,)
                    )  # type:ignore[union-attr]
                    self.running_var.materialize(
                        (self.num_features,)
                    )  # type:ignore[union-attr]
                self.reset_parameters()


class InstanceNormMixin:
    """Skips dimension check."""

    def __init__(self, *args, affine=True, **kwargs):
        self.no_batch_dim = None  # no_batch_dim has to be set at runtime
        super().__init__(*args, affine=affine, **kwargs)

    def set_no_dim_batch_dim(self, no_batch_dim):
        self.no_batch_dim = no_batch_dim

    def _check_input_dim(self, input):
        return

    def _get_no_batch_dim(self):
        return self.no_batch_dim


class LazyInstanceNormUnsafe(InstanceNormMixin, _LazyInstanceNorm):
    pass


class InstanceNormUnsafe(InstanceNormMixin, _InstanceNorm):
    pass


class InstanceNormWrapper(torch.nn.Module):
    def __init__(self, torch_params, *args, affine=True, **kwargs):
        super().__init__()
        self.has_lazy = len(torch_params) == 0
        if self.has_lazy:
            self.inu = LazyInstanceNormUnsafe(*args, affine=affine, **kwargs)
        else:
            kwargs["num_features"] = torch_params[0].shape[0]
            self.inu = InstanceNormUnsafe(*args, affine=affine, **kwargs)
            keys = ["weight", "bias"]
            for key, value in zip(keys, torch_params):
                getattr(self.inu, key).data = value

    def forward(self, input, scale=None, B=None):
        if self.has_lazy:
            self.inu.initialize_parameters(input)

        if scale is not None:
            getattr(self.inu, "weight").data = scale
        if B is not None:
            getattr(self.inu, "bias").data = B

        if self.inu.no_batch_dim is None:
            self.inu.set_no_dim_batch_dim(input.dim() - 1)

        return self.inu.forward(input)



import torch.nn.functional as F

from onnx2pytorch.operations.base import Operator


class Pad(Operator):
    def __init__(self, mode="constant", padding=None):
        self.mode = mode
        self.padding = padding
        super().__init__()

    def forward(self, input, pads=None, value=0):
        if self.padding is not None:
            pads = self.padding
        elif pads is None:
            raise TypeError("forward() missing 1 required positional argument: 'pads'")
        out = F.pad(input, list(pads), mode=self.mode, value=value)
        return out

    def extra_repr(self) -> str:
        return "mode={}, padding={}".format(self.mode, self.padding)
